
/*
 * File Z3Interfacing.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J8. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J8 only because so is required by the CASC-J8
 * rules, and for no other reason.
 */


#ifndef __Z3Interfacing__
#define __Z3Interfacing__

#if VZ3

#include "Lib/DHMap.hpp"

#include "SATSolver.hpp"
#include "SATLiteral.hpp"
#include "SATClause.hpp"
#include "SATInference.hpp"
#include "SAT2FO.hpp"

#include "z3++.h"
#include "z3_api.h"

namespace SAT{

class Z3Interfacing : public PrimitiveProofRecordingSATSolver
{
public: 
  CLASS_NAME(Z3Interfacing);
  USE_ALLOCATOR(Z3Interfacing);
  
 
  Z3Interfacing(const Shell::Options& opts, SAT2FO& s2f, bool unsatCoresForAssumptions = false);

  void addClause(SATClause* cl) override;

  virtual Status solve(unsigned conflictCountLimit) override;
 
  virtual VarAssignment getAssignment(unsigned var) override;

 
  virtual bool isZeroImplied(unsigned var) override;
 
  virtual void collectZeroImplied(SATLiteralStack& acc) override;
 
  virtual SATClause* getZeroImpliedCertificate(unsigned var) override;

  
  virtual void ensureVarCount(unsigned newVarCnt) override {
    CALL("Z3Interfacing::ensureVarCnt");
    _varCnt = max(newVarCnt,_varCnt);
  }

  virtual unsigned newVar() override {
    CALL("Z3Interfacing::newVar");
    return ++_varCnt;
  }

  
  virtual void suggestPolarity(unsigned var, unsigned pol) override {}
  
  virtual void addAssumption(SATLiteral lit) override;
  virtual void retractAllAssumptions() override { _assumptions.resize(0); }
  virtual bool hasAssumptions() const override { return !_assumptions.empty(); }

  virtual Status solveUnderAssumptions(const SATLiteralStack& assumps, unsigned, bool) override;


  virtual void recordSource(unsigned satlitvar, Literal* lit) override {
    
  };
  
 
  SATClauseList* getRefutationPremiseList() override{ return 0; } 

  SATClause* getRefutation() override;  

private:
  
  unsigned _varCnt;

  
  SAT2FO& sat2fo;

  
  z3::sort getz3sort(unsigned s);

  
  z3::expr to_int(z3::expr e) {
        return z3::expr(e.ctx(), Z3_mk_real2int(e.ctx(), e));
  }
  z3::expr to_real(z3::expr e) {
        return z3::expr(e.ctx(), Z3_mk_int2real(e.ctx(), e));
  }
  z3::expr ceiling(z3::expr e){
        return -to_real(to_int((e)));
  }
  z3::expr is_even(z3::expr e) {
        z3::context& ctx = e.ctx();
        z3::expr two = ctx.int_val(2);
        z3::expr m = z3::expr(ctx, Z3_mk_mod(ctx, e, two));
        return m == 0;
  }

  z3::expr truncate(z3::expr e) {
        return ite(e >= 0, to_int(e), ceiling(e));
  }

  void addTruncatedOperations(z3::expr_vector, Interpretation qi, Interpretation ti, unsigned srt);
  void addFloorOperations(z3::expr_vector, Interpretation qi, Interpretation ti, unsigned srt);

public:
  z3::expr getz3expr(Term* trm,bool islit,bool&nameExpression);
private:
  z3::expr getRepresentation(SATLiteral lit);

  Status _status;
  z3::context _context;
  z3::solver _solver;
  z3::model _model;

  z3::expr_vector _assumptions;
  bool _unsatCoreForAssumptions;

  bool _showZ3;
  bool _unsatCoreForRefutations;

  DHSet<unsigned> _namedExpressions; 
  z3::expr getNameExpr(unsigned var){
    vstring name = "v"+Lib::Int::toString(var);
    return  _context.bool_const(name.c_str());
  }

};

}

#endif
#endif
